# -*- coding: utf8 -*-
__all__ = ('Flags',)
import struct


class Flags(object):
    """
    Convience class for handling bit flags.
    """
    def __init__(self, binary_format, flags):
        object.__setattr__(self, 'binary_format', binary_format)
        object.__setattr__(self, 'flags', flags)
        object.__setattr__(self, '_value', 0)
        object.__setattr__(self, '_cache', struct.Struct(binary_format))

    def pack(self):
        """
        A shortcut for `struct.pack(flag.binary_format, flag.value)`.
        """
        return self._cache.pack(self.value)

    @property
    def value(self):
        """
        The numeric value of the bitfield.
        """
        return self._value

    def unpack(self, source):
        """
        A shortcut for `struct.unpack(flag.binary_format, <bytes>)`.
        """
        self._value = self._cache.unpack(source)[0]

    def get(self, name):
        """
        Returns the value of the field `name`.
        """
        return bool(self.flags[name] & self.value)

    def set(self, name, value):
        """
        Sets the value of the field `name` to `value`, which is `True` or
        `False`.
        """
        flag = self.flags[name]
        self._value = (self.value | flag) if value else (self.value & ~flag)

    def __getattr__(self, attr):
        if attr not in self.flags:
            return object.__getattr__(self, attr)
        return self.get(attr)

    def __setattr__(self, attr, value):
        if attr not in self.flags:
            return object.__setattr__(self, attr, value)
        self.set(attr, value)

    def to_dict(self):
        """
        Returns this `Flags` object's fields as a dictionary.
        """
        return dict((k, self.get(k)) for k in self.flags.keys())



# -*- coding: utf-8 -*-
import click

from jawa.cf import ClassFile


@click.group()
def dism():
    pass


@dism.command(help=(
    'A mostly-compatible clone of the javap disassembler.'
))
@click.option('-v', '-verbose', is_flag=True)
@click.option(
    '-l',
    'tables',
    is_flag=True,
    help='Print line number and local variable tables'
)
@click.option(
    '-public',
    'scope',
    flag_value='public'
)
@click.option(
    '-protected',
    'scope',
    flag_value='protected'
)
@click.option(
    '-package',
    'scope',
    flag_value='package',
    default=True
)
@click.option(
    '-p',
    '-private',
    'scope',
    flag_value='private'
)
@click.option(
    '-c',
    'disassemble',
    is_flag=True
)
@click.option(
    '-s',
    is_flag=True
)
@click.option(
    '-constants',
    is_flag=True
)
@click.option(
    '-classpath',
    '-cp',
    type=click.Path(resolve_path=True)
)
@click.option(
    '-bootclasspath',
    type=click.Path(resolve_path=True)
)
@click.argument('classes', nargs=-1)
def javap(**args):
    for class_ in args['classes']:
        with open(class_, 'rb') as fin:
            cf = ClassFile(fin)
            _javap(cf, args)


def _javap(cf, args):
    flags = cf.access_flags.to_dict()
    flags = [k[4:] for k, v in flags.items() if v]

    click.echo(
        '{flags} {cf.this.name.value} '.format(
            cf=cf,
            flags=' '.join(flags)
        ),
        nl=False
    )
    if cf.super_.name.value != 'java/lang/Object':
        click.echo('extends {cf.super_.name.value} '.format(
            cf=cf
        ), nl=False)

    if cf.interfaces:
        click.echo('{what} {ifs} '.format(
            what='extends' if cf.access_flags.acc_interface else 'implements',
            ifs=', '.join(c.name.value for c in cf.interfaces)
        ), nl=False)

    click.echo('{')

    click.echo('}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import click

from jawa.cli.disassemble import dism
from jawa.cli.info import info


@click.group()
def cli():
    pass

cli.add_command(dism)
cli.add_command(info)



# -*- coding: utf-8 -*-
import json

import click

from jawa.df import DexFile


@click.group()
def info():
    pass


@info.command(help='Display header information for DEX files.')
@click.argument('input', type=click.File('rb'))
def dex(input):
    df = DexFile(input)
    op = {
        'format': {
            'version': df.version,
            'little_endian': df.little_endian,
        },
        'verification': {
            'signature': ''.join(
                format(b, '02X') for b in
                df.signature
            ),
            'checksum': df.checksum,
        },
        'link_segment': {
            'offset': df.link_off,
            'size': df.link_size
        },
        'map_segment': {
            'offset': df.map_off
        }
    }

    click.echo(json.dumps(op, sort_keys=True, indent=2))



# -*- coding: utf8 -*-
__all__ = (
    'Instruction',
    'Operand',
    'OperandTypes',
    'read_instruction',
    'write_instruction',
    'opcode_table',
    'definition_from_mnemonic',
    'definition_from_opcode'
)

import struct

from itertools import repeat
from collections import namedtuple

Operand = namedtuple('Operand', ['op_type', 'value'])
_Instruction = namedtuple('Instruction', [
    'mnemonic',
    'opcode',
    'operands',
    'pos'
])

# Opcodes that can be prefixed by the `wide` opcode.
_wide = (
    0x15, 0x17, 0x19, 0x16,
    0x18, 0x36, 0x38, 0x3A,
    0x37, 0x39, 0xA9, 0x84
)


class Instruction(_Instruction):
    """
    Represents a single JVM instruction, consisting
    of an opcode and its potential operands.
    """
    __slots__ = ()

    @classmethod
    def from_mnemonic(cls, mnemonic, operands=None):
        return cls(
            mnemonic,
            _opcode_by_opname[mnemonic][0],
            operands or [],
            0
        )

    @classmethod
    def from_opcode(cls, opcode, operands=None):
        return cls(
            opcode_table[opcode][0],
            opcode,
            operands or [],
            0
        )

    def size_on_disk(self, start_pos=0):
        """
        Returns the size of this instruction and its operands when
        packed. `start_pos` is required for the `tableswitch` and
        `lookupswitch` instruction as the padding depends on alignment.
        """
        size, fmts = 1, definition_from_opcode(self.opcode)[1]
        if self.wide:
            size += 2
            # Special case for iinc which has a 2nd extended operand.
            if self.opcode == 0x84:
                size += 2
        elif fmts:
            # A simple opcode with simple operands.
            for fmt, _ in fmts:
                size += fmt.size
        elif self.opcode == 0xAB:
            # lookupswitch
            padding = 4 - (start_pos + 1) % 4
            padding = padding if padding != 4 else 0
            size += padding
            # default & npairs
            size += 8
            size += len(self.operands[0]) * 8
        elif self.opcode == 0xAA:
            # tableswitch
            raise NotImplementedError()

        return size

    @property
    def wide(self):
        """
        ``True`` if this instruction needs to be prefixed by the WIDE
        opcode.
        """
        if self.opcode not in _wide:
            return False

        if self.operands[0].value >= 255:
            return True

        if self.opcode == 0x84:
            if self.operands[1].value >= 255:
                return True

        return False


class OperandTypes(object):
    """
    Constants used to determine the "type" of operand on an opcode,
    such as a BRANCH [offset] or a LITERAL [value].
    """
    LITERAL = 10
    LOCAL_INDEX = 20
    CONSTANT_INDEX = 30
    BRANCH = 40
    PADDING = 50


ubyte = struct.Struct('>B')
ushort = struct.Struct('>H')
byte = struct.Struct('>b')
short = struct.Struct('>h')
integer = struct.Struct('>i')

#: An opcode to mnemonic & operand format mapping for every opcode
#: supported by the JVM.
opcode_table = {
    0x32: ('aaload', None),
    0x53: ('aastore', None),
    0x01: ('aconst_null', None),
    0x19: ('aload', [(ubyte, 20)]),
    0x2A: ('aload_0', None),
    0x2B: ('aload_1', None),
    0x2C: ('aload_2', None),
    0x2D: ('aload_3', None),
    0xBD: ('anewarray', [(ushort, 10)]),
    0xB0: ('areturn', None),
    0xBE: ('arraylength', None),
    0x3A: ('astore', [(ubyte, 20)]),
    0x4B: ('astore_0', None),
    0x4C: ('astore_1', None),
    0x4D: ('astore_2', None),
    0x4E: ('astore_3', None),
    0xBF: ('athrow', None),
    0x33: ('baload', None),
    0x54: ('bastore', None),
    0x10: ('bipush', [(ubyte, 10)]),
    0x34: ('caload', None),
    0x55: ('castore', None),
    0xC0: ('checkcast', [(ushort, 30)]),
    0x90: ('d2f', None),
    0x8E: ('d2i', None),
    0x8F: ('d2l', None),
    0x63: ('dadd', None),
    0x31: ('daload', None),
    0x52: ('dastore', None),
    0x98: ('dcmpg', None),
    0x97: ('dcmpl', None),
    0x0E: ('dconst_0', None),
    0x0F: ('dconst_1', None),
    0x6F: ('ddiv', None),
    0x18: ('dload', [(ubyte, 20)]),
    0x26: ('dload_0', None),
    0x27: ('dload_1', None),
    0x28: ('dload_2', None),
    0x29: ('dload_3', None),
    0x6B: ('dmul', None),
    0x77: ('dneg', None),
    0x73: ('drem', None),
    0xAF: ('dreturn', None),
    0x39: ('dstore', [(ubyte, 20)]),
    0x47: ('dstore_0', None),
    0x48: ('dstore_1', None),
    0x49: ('dstore_2', None),
    0x4a: ('dstore_3', None),
    0x67: ('dsub', None),
    0x59: ('dup', None),
    0x5A: ('dup_x1', None),
    0x5B: ('dup_x2', None),
    0x5C: ('dup2', None),
    0x5D: ('dup2_x1', None),
    0x5E: ('dup2_x2', None),
    0x8D: ('f2d', None),
    0x8B: ('f2i', None),
    0x8C: ('f2l', None),
    0x62: ('fadd', None),
    0x30: ('faload', None),
    0x51: ('fastore', None),
    0x96: ('fcmpg', None),
    0x95: ('fcmpl', None),
    0x0B: ('fconst_0', None),
    0x0C: ('fconst_1', None),
    0x0D: ('fconst_2', None),
    0x6E: ('fdiv', None),
    0x17: ('fload', [(ubyte, 20)]),
    0x22: ('fload_0', None),
    0x23: ('fload_1', None),
    0x24: ('fload_2', None),
    0x25: ('fload_3', None),
    0x6A: ('fmul', None),
    0x76: ('fneg', None),
    0x72: ('frem', None),
    0xAE: ('freturn', None),
    0x38: ('fstore', [(ubyte, 20)]),
    0x43: ('fstore_0', None),
    0x44: ('fstore_1', None),
    0x45: ('fstore_2', None),
    0x46: ('fstore_3', None),
    0x66: ('fsub', None),
    0xB4: ('getfield', [(ushort, 30)]),
    0xB2: ('getstatic', [(ushort, 30)]),
    0xA7: ('goto', [(short, 40)]),
    0xC8: ('goto_w', [(integer, 40)]),
    0x91: ('ib2', None),
    0x92: ('i2c', None),
    0x87: ('i2d', None),
    0x86: ('i2f', None),
    0x85: ('i2l', None),
    0x93: ('i2s', None),
    0x60: ('iadd', None),
    0x2E: ('iaload', None),
    0x7E: ('iand', None),
    0x4F: ('iastore', None),
    0x02: ('iconst_m1', None),
    0x03: ('iconst_0', None),
    0x04: ('iconst_1', None),
    0x05: ('iconst_2', None),
    0x06: ('iconst_3', None),
    0x07: ('iconst_4', None),
    0x08: ('iconst_5', None),
    0x6C: ('idiv', None),
    0xA5: ('if_acmpeq', [(short, 40)]),
    0xA6: ('if_acmpne', [(short, 40)]),
    0x9F: ('if_icmpeq', [(short, 40)]),
    0xA0: ('if_icmpne', [(short, 40)]),
    0xA1: ('if_icmplt', [(short, 40)]),
    0xA2: ('if_icmpge', [(short, 40)]),
    0xA3: ('if_icmpgt', [(short, 40)]),
    0xA4: ('if_icmple', [(short, 40)]),
    0x99: ('ifeq', [(short, 40)]),
    0x9A: ('ifne', [(short, 40)]),
    0x9B: ('iflt', [(short, 40)]),
    0x9C: ('ifge', [(short, 40)]),
    0x9D: ('ifgt', [(short, 40)]),
    0x9E: ('ifle', [(short, 40)]),
    0xC7: ('ifnonnull', [(short, 40)]),
    0xC6: ('ifnull', [(short, 40)]),
    0x84: ('iinc', [(ubyte, 20), (ubyte, 10)]),
    0x15: ('iload', [(ubyte, 20)]),
    0x1A: ('iload_0', None),
    0x1B: ('iload_1', None),
    0x1C: ('iload_2', None),
    0x1D: ('iload_3', None),
    0x68: ('imul', None),
    0x74: ('ineg', None),
    0xC1: ('instanceof', [(ushort, 30)]),
    0xB9: ('invokeinterface', [(ushort, 30), (ubyte, 10), (ubyte, 50)]),
    0xB7: ('invokespecial', [(ushort, 30)]),
    0xB8: ('invokestatic', [(ushort, 30)]),
    0xB6: ('invokevirtual', [(ushort, 30)]),
    0x80: ('ior', None),
    0x70: ('irem', None),
    0xAC: ('ireturn', None),
    0x78: ('ishl', None),
    0x7A: ('ishr', None),
    0x36: ('istore', [(ubyte, 20)]),
    0x3B: ('istore_0', None),
    0x3C: ('istore_1', None),
    0x3D: ('istore_2', None),
    0x3E: ('istore_3', None),
    0x64: ('isub', None),
    0x7C: ('iushr', None),
    0x82: ('ixor', None),
    0xA8: ('jsr', [(short, 40)]),
    0xC9: ('jsr_w', [(integer, 40)]),
    0x8A: ('l2d', None),
    0x89: ('l2f', None),
    0x88: ('l2i', None),
    0x61: ('ladd', None),
    0x2F: ('laload', None),
    0x7F: ('land', None),
    0x50: ('lastore', None),
    0x94: ('lcmp', None),
    0x09: ('lconst_0', None),
    0x0A: ('lconst_1', None),
    0x12: ('ldc', [(ubyte, 30)]),
    0x13: ('ldc_w', [(ushort, 30)]),
    0x14: ('ldc2_w', [(ushort, 30)]),
    0x6D: ('ldiv', None),
    0x16: ('lload', [(ubyte, 20)]),
    0x1E: ('lload_0', None),
    0x1F: ('lload_1', None),
    0x20: ('lload_2', None),
    0x21: ('lload_3', None),
    0x69: ('lmul', None),
    0x75: ('lneg', None),
    0xAB: ('lookupswitch', None),
    0x81: ('lor', None),
    0x71: ('lrem', None),
    0xAD: ('lreturn', None),
    0x79: ('lshl', None),
    0x7B: ('lshr', None),
    0x37: ('lstore', [(ubyte, 20)]),
    0x3F: ('lstore_0', None),
    0x40: ('lstore_1', None),
    0x41: ('lstore_2', None),
    0x42: ('lstore_3', None),
    0x65: ('lsub', None),
    0x7D: ('lushr', None),
    0x83: ('lxor', None),
    0xC2: ('monitorenter', None),
    0xC3: ('monitorexit', None),
    0xC5: ('multianewarray', [(ushort, 30), (ubyte, 10)]),
    0xBB: ('new', [(ushort, 30)]),
    0xBC: ('newarray', [(ubyte, 10)]),
    0x00: ('nop', None),
    0x57: ('pop', None),
    0x58: ('pop2', None),
    0xB5: ('putfield', [(ushort, 30)]),
    0xB3: ('putstatic', [(ushort, 30)]),
    0xA9: ('ret', [(ubyte, 20)]),
    0xB1: ('return', None),
    0x35: ('saload', None),
    0x56: ('sastore', None),
    0x11: ('sipush', [(ushort, 10)]),
    0x5F: ('swap', None),
    0xAA: ('tableswitch', None),
    0xC4: ('wide', None),
    0xCA: ('breakpoint', None),
    0xFE: ('impdep1', None),
    0xFF: ('impdep2', None)
}

# Creates a new dict with operand and mnemonic swapped for assembling.
_opcode_by_opname = dict((o[0], (k,) + o[1:]) for k, o in opcode_table.items())


def write_instruction(fout, start_pos, ins):
    """
    Writes a single instruction of `opcode` with `operands` to `fout`.

    :param fout: Any file-like object providing ``write()``.
    :param start_pos: The current position in the stream.
    :param ins: The `Instruction` to write.
    """
    opcode, operands = ins.opcode, ins.operands
    fmt_operands = definition_from_opcode(opcode)[1]

    if ins.wide:
        # The "WIDE" prefix
        fout.write(ubyte.pack(0xC4))
        # The real opcode.
        fout.write(ubyte.pack(opcode))
        fout.write(ushort.pack(operands[0].value))
        if opcode == 0x84:
            fout.write(short.pack(operands[1].value))
    # A normal simple opcode with simple operands.
    elif fmt_operands:
        fout.write(ubyte.pack(opcode))
        for i, (fmt, _) in enumerate(fmt_operands):
            fout.write(fmt.pack(operands[i].value))
    # Special case for lookupswitch, which also has a unique.
    elif opcode == 0xAB:
        fout.write(ubyte.pack(opcode))
        # assemble([
        #     ('lookupswitch', {
        #         2: -3,
        #         4: 5
        #     }, <default>)
        # ])
        padding = 4 - (start_pos + 1) % 4
        padding = padding if padding != 4 else 0
        fout.write(struct.pack('{0}x'.format(padding)))
        fout.write(struct.pack('>ii',
            operands[1].value,
            len(operands[0])
        ))
        for key in sorted(operands[0].keys()):
            fout.write(struct.pack('>ii',
                key,
                operands[0][key].value
            ))
    # Special case for tableswitch.
    elif opcode == 0xAA:
        raise NotImplementedError()
    else:
        # opcode with no operands.
        fout.write(ubyte.pack(opcode))


def read_instruction(fio, start_pos):
    """
    Reads a single instruction from `fio` and returns it, or ``None`` if
    the stream is empty.

    :param fio: Any file-like object providing ``read()``.
    :param start_pos: The current position in the stream.
    """
    try:
        op = ubyte.unpack(fio.read(1))[0]
    except struct.error:
        return None

    name, operands = opcode_table[op][:2]

    final_operands = []
    # Most opcodes have simple operands.
    if operands:
        for fmt, type_ in operands:
            final_operands.append(
                Operand(type_, fmt.unpack(fio.read(fmt.size))[0]))
    # Special case for lookupswitch.
    elif op == 0xAB:
        # Get rid of the alignment padding.
        padding = 4 - (start_pos + 1) % 4
        padding = padding if padding != 4 else 0
        fio.read(padding)

        # Default branch address and branch count.
        default, npairs = struct.unpack('>ii', fio.read(8))

        pairs = {}
        for _ in repeat(None, npairs):
            match, offset = struct.unpack('>ii', fio.read(8))
            pairs[match] = offset

        final_operands.append(pairs)
        final_operands.append(Operand(OperandTypes.BRANCH, default))
    # Special case for tableswitch
    elif op == 0xAA:
        # Get rid of the alignment padding.
        padding = 4 - (start_pos + 1) % 4
        padding = padding if padding != 4 else 0
        fio.read(padding)

        default, low, high = struct.unpack('>iii', fio.read(12))
        final_operands.append(Operand(OperandTypes.BRANCH, default))
        final_operands.append(Operand(OperandTypes.LITERAL, low))
        final_operands.append(Operand(OperandTypes.LITERAL, high))

        for _ in repeat(None, high - low + 1):
            offset = struct.unpack('>i', fio.read(4))[0]
            final_operands.append(Operand(OperandTypes.BRANCH, offset))
    # Special case for the wide prefix
    elif op == 0xC4:
        real_op = ubyte.unpack(fio.read(1))[0]
        name, operands = opcode_table[real_op][:2]
        final_operands.append(Operand(
            OperandTypes.LOCAL_INDEX,
            ushort.unpack(fio.read(2))[0]
        ))
        # Further special case for iinc.
        if real_op == 0x84:
            final_operands.append(Operand(
                OperandTypes.LITERAL,
                short.unpack(fio.read(2))[0]
            ))

    return Instruction(name, op, final_operands, start_pos)


def definition_from_mnemonic(mnemonic):
    """
    Returns the definition of an instruction by its mnemonic in the
    form: ``(opcode, operand_fmt)``
    """
    return _opcode_by_opname[mnemonic]


def definition_from_opcode(opcode):
    """
    Returns the definition of an instruction by its opcode in the
    form: ``(mnemonic, operand_fmt)``
    """
    return opcode_table[opcode]



# -*- coding: utf8 -*-
__all__ = ('FieldTable', 'Field')

from struct import unpack, pack
from itertools import repeat

from jawa.util.flags import Flags
from jawa.attribute import AttributeTable
from jawa.attributes.constant_value import ConstantValueAttribute


class Field(object):
    def __init__(self, cf):
        self._cf = cf
        self._access_flags = Flags('>H', {
            'acc_public': 0x0001,
            'acc_private': 0x0002,
            'acc_protected': 0x0004,
            'acc_static': 0x0008,
            'acc_final': 0x0010,
            'acc_volatile': 0x0040,
            'acc_transient': 0x0080,
            'acc_synthetic': 0x1000,
            'acc_enum': 0x4000
        })
        self._name_index = 0
        self._descriptor_index = 0
        self._attributes = AttributeTable(cf)

    @property
    def descriptor(self):
        return self._cf.constants[self._descriptor_index]

    @property
    def name(self):
        return self._cf.constants[self._name_index]

    @property
    def access_flags(self):
        return self._access_flags

    @property
    def attributes(self):
        return self._attributes

    @property
    def value(self):
        """
        A shortcut for the field's ConstantValue attribute, should one exist.
        """
        constant_value = self.attributes.find_one(name='ConstantValue')
        return constant_value

    def unpack(self, fio):
        """
        Read the Field from the file-like object `fio`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param fio: Any file-like object providing `read()`
        """
        self.access_flags.unpack(fio.read(2))
        self._name_index, self._descriptor_index = unpack('>HH', fio.read(4))
        self._attributes.unpack(fio)

    def pack(self, fout):
        """
        Write the Field to the file-like object `fout`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be calle=d for you when saving a ClassFile.

        :param fout: Any file-like object providing `write()`
        """
        fout.write(self.access_flags.pack())
        fout.write(pack('>HH', self._name_index, self._descriptor_index))
        self._attributes.pack(fout)


class FieldTable(object):
    def __init__(self, cf):
        self._cf = cf
        self._table = []

    def append(self, field):
        self._table.append(field)

    def find_and_remove(self, f):
        """
        Removes any and all fields for which `f(field)` returns `True`.
        """
        self._table = [fld for fld in self._table if not f(fld)]

    def remove(self, field):
        """
        Removes a `Field` from the table by identity.
        """
        self._table = [fld for fld in self._table if fld is not field]

    def create(self, name, descriptor):
        """
        Creates a new field from `name` and `descriptor`. For example::

            >>> from jawa import ClassFile
            >>> cf = ClassFile.create('BeerCounter')
            >>> field = cf.fields.create('BeerCount', 'I')

        """
        field = Field(self._cf)
        name = self._cf.constants.create_utf8(name)
        descriptor = self._cf.constants.create_utf8(descriptor)
        field._name_index = name.index
        field._descriptor_index = descriptor.index
        field.access_flags.acc_public = True
        self.append(field)
        return field

    def create_static(self, name, descriptor, value):
        """
        A shortcut for creating a static field with a
        :class:`~jawa.attributes.constant_value.ConstantValueAttribute`
        set to the :class:`~jawa.constants.Constant` `value`. For example,
        to create a string with the classic "Hello World!"::

            >>> from jawa import ClassFile
            >>> cf = ClassFile.create('BeerCounter')
            >>> field = cf.fields.create_static(
            ...    'HelloWorld',
            ...    'Ljava/lang/String;',
            ...    cf.constants.create_string('Hello World!')
            ... )
        """
        field = self.create(name, descriptor)
        field.attributes.create(ConstantValueAttribute, value)
        field.access_flags.acc_static = True
        return field

    def __iter__(self):
        for field in self._table:
            yield field

    def unpack(self, fio):
        """
        Read the FieldTable from the file-like object `fio`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param fio: Any file-like object providing `read()`
        """
        field_count = unpack('>H', fio.read(2))[0]
        for _ in repeat(None, field_count):
            field = Field(self._cf)
            field.unpack(fio)
            self.append(field)

    def pack(self, fout):
        """
        Write the FieldTable to the file-like object `fout`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be calle=d for you when saving a ClassFile.

        :param fout: Any file-like object providing `write()`
        """
        fout.write(pack('>H', len(self)))
        for field in self._table:
            field.pack(fout)

    def __len__(self):
        return len(self._table)

    def find(self, name=None, type_=None, f=None):
        """
        Iterates over the fields table, yielding each matching method. Calling
        without any arguments is equivelent to iterating over the table.

        :param name: The name of the field(s) to find.
        :param type_: The field descriptor (Ex: 'I')
        :param f: Any callable which takes one argument (the field).
        """
        for field in self._table:
            if name is not None and field.name.value != name:
                continue

            descriptor = field.descriptor.value
            if type_ is not None and type_ != descriptor:
                continue

            if f is not None and not f(field):
                continue

            yield field

    def find_one(self, *args, **kwargs):
        """
        Same as ``find()`` but returns only the first result, or `None` if
        nothing was found.
        """
        try:
            return next(self.find(*args, **kwargs))
        except StopIteration:
            return None



# -*- coding: utf8 -*-
"""
ClassFile support.

The :mod:`jawa.cf` module provides tools for working with JVM ``.class``
ClassFiles.
"""
from struct import pack, unpack
from collections import namedtuple


from jawa.constants import ConstantPool
from jawa.fields import FieldTable
from jawa.methods import MethodTable
from jawa.attribute import AttributeTable
from jawa.util.flags import Flags


class ClassVersion(namedtuple('ClassVersion', ['major', 'minor'])):
    """ClassFile file format version."""

    __slots__ = ()

    @property
    def human(self):
        """
        A human-readable string identifying this version.

        If the version is unknown, `None` is returned instead.
        """
        return {
            0x33: 'J2SE_7',
            0x32: 'J2SE_6',
            0x31: 'J2SE_5',
            0x30: 'JDK1_4',
            0x2F: 'JDK1_3',
            0x2E: 'JDK1_2',
            0x2D: 'JDK1_1',
        }.get(self.major, None)


class ClassFile(object):
    """
    Implements the JVM ClassFile (files typically ending in ``.class``).

    To open an existing ClassFile::

        from jawa import ClassFile
        with open('HelloWorld.class', 'rb') as fin:
            cf = ClassFile(fin)

    To save a newly created or modified ClassFile::

        with open('HelloWorld.class', 'wb') as fout:
            cf.save(fout)

    To create a new ClassFile, use the helper :meth:`~ClassFile.create`::

        from jawa import ClassFile
        cf = ClassFile.create('HelloWorld')
        with open('HelloWorld.class', 'wb') as fout:
            cf.save(fout)

    :meth:`~ClassFile.create` sets up some reasonable defaults equivelent to:

    .. code-block:: java

        public class HelloWorld extends java.lang.Object{
        }

    :param fio: any file-like object providing ``.read()``.
    """

    #: The JVM ClassFile magic number.
    MAGIC = 0xCAFEBABE

    def __init__(self, fio=None):
        # Default to J2SE_7
        self._version = ClassVersion(0x32, 0)
        self._constants = ConstantPool()
        self._access_flags = Flags('>H', {
            'acc_public': 0x0001,
            'acc_final': 0x0010,
            'acc_super': 0x0020,
            'acc_interface': 0x0200,
            'acc_abstract': 0x0400,
            'acc_synthetic': 0x1000,
            'acc_annotation': 0x2000,
            'acc_enum': 0x4000
        })
        self._this = 0
        self._super = 0
        self._interfaces = []
        self._fields = FieldTable(self)
        self._methods = MethodTable(self)
        self._attributes = AttributeTable(self)

        if fio:
            self._from_io(fio)

    @classmethod
    def create(cls, this, super_='java/lang/Object'):
        """
        A utility which sets up reasonable defaults for a new public class.

        :param this: The name of this class.
        :param super_: The name of this class's superclass.
        """
        cf = ClassFile()
        cf.access_flags.acc_public = True
        cf.access_flags.acc_super = True

        cf.this = cf.constants.create_class(this)
        cf.super_ = cf.constants.create_class(super_)

        return cf

    def save(self, fout):
        """
        Saves the class to the file-like object `fout`.
        """
        write = fout.write

        write(pack(
            '>IHH',
            ClassFile.MAGIC,
            self.version.minor,
            self.version.major
        ))

        self._constants.pack(fout)

        write(self.access_flags.pack())
        write(pack(
            '>HHH{0}H'.format(len(self._interfaces)),
            self._this,
            self._super,
            len(self._interfaces),
            *self._interfaces
        ))

        self._fields.pack(fout)
        self._methods.pack(fout)
        self._attributes.pack(fout)

    def _from_io(self, fio):
        """
        Loads an existing JVM ClassFile from any file-like object.
        """
        read = fio.read

        if unpack('>I', fio.read(4))[0] != ClassFile.MAGIC:
            raise ValueError('invalid magic number')

        # The version is swapped on disk to (minor, major), so swap it back.
        self.version = unpack('>HH', fio.read(4))[::-1]

        self._constants.unpack(fio)

        # ClassFile access_flags, see section #4.1 of the JVM specs.
        self.access_flags.unpack(read(2))

        # The CONSTANT_Class indexes for "this" class and its superclass.
        # Interfaces are a simple list of CONSTANT_Class indexes.
        self._this, self._super, interfaces_count = unpack('>HHH', read(6))
        self._interfaces = unpack(
            '>{0}H'.format(interfaces_count),
            read(2 * interfaces_count)
        )

        self._fields.unpack(fio)
        self._methods.unpack(fio)
        self._attributes.unpack(fio)

    @property
    def version(self):
        """
        The :class:`~jawa.cf.ClassVersion` for this class.

        Example::

            >>> cf = ClassFile.create('HelloWorld')
            >>> cf.version = 51, 0
            >>> print(cf.version)
            ClassVersion(major=51, minor=0)
            >>> print(cf.version.major)
            51
        """
        return self._version

    @version.setter
    def version(self, (major, minor)):
        self._version = ClassVersion(major, minor)

    @property
    def constants(self):
        """
        The :class:`~jawa.cp.ConstantPool` for this class.
        """
        return self._constants

    @property
    def access_flags(self):
        return self._access_flags

    @property
    def this(self):
        """
        The :class:`~jawa.constants.ConstantClass` which represents this class.
        """
        return self.constants.get(self._this)

    @this.setter
    def this(self, value):
        self._this = value.index

    @property
    def super_(self):
        """
        The :class:`~jawa.constants.ConstantClass` which represents this
        class's superclass.
        """
        return self.constants.get(self._super)

    @super_.setter
    def super_(self, value):
        self._super = value.index

    @property
    def interfaces(self):
        """
        A list of direct superinterfaces of this class as indexes into
        the constant pool, in left-to-right order.
        """
        for interface_index in self._interfaces:
            yield self._constants[interface_index]

    @property
    def fields(self):
        """
        The :class:`~jawa.fields.FieldTable` for this class.
        """
        return self._fields

    @property
    def methods(self):
        """
        The :class:`~jawa.methods.MethodTable` for this class.
        """
        return self._methods

    @property
    def attributes(self):
        """
        The :class:`~jawa.attribute.AttributeTable` for this class.
        """
        return self._attributes



# -*- coding: utf8 -*-
from struct import unpack_from, pack
from jawa.attribute import Attribute


class SourceFileAttribute(Attribute):
    ADDED_IN = '1.0.2'
    MINIMUM_CLASS_VERSION = (45, 3)

    def __init__(self, table, sourcefile=None, name_index=None):
        super(SourceFileAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'SourceFile'
            ).index
        )
        self._sourcefile_index = sourcefile.index if sourcefile else None

    def unpack(self, info):
        self._sourcefile_index = unpack_from('>H', info)[0]

    @property
    def info(self):
        return pack('>H', self._sourcefile_index)

    @property
    def sourcefile(self):
        return self._cf.constants[self._sourcefile_index]



# -*- coding: utf8 -*-
from collections import namedtuple

from jawa.constants import Constant
from jawa.util.bytecode import (
    Operand,
    OperandTypes,
    Instruction,
    definition_from_mnemonic
)


Label = namedtuple('Label', ['name'])


def assemble(code):
    """
    Assemble the given iterable of mnemonics, operands, and lables.

    A convienience over constructing individual Instruction and Operand
    objects, the output of this function can be directly piped to
    :class:`~jawa.attributes.code.CodeAttribute.assemble()` to produce
    executable bytecode.

    As a simple example, lets produce an infinite loop:

        >>> from jawa.assemble import assemble, Label
        >>> print(list(assemble((
        ...     Label('start'),
        ...     ('goto', Label('start'))
        ... ))))
        [Instruction(mnemonic='goto', opcode=167, operands=[
            Operand(op_type=40, value=0)], pos=0)]

    For a more complex example, see examples/hello_world.py.
    """
    final = []

    # We need to make three passes, because we cannot know the offset for
    # jump labels until after we've figured out the PC for each instructions,
    # which is complicated by the variable-width instructions set and
    # alignment padding.
    for line in code:
        if isinstance(line, Label):
            final.append(line)
            continue

        mnemonic, operands = line[0], line[1:]
        operand_fmts = definition_from_mnemonic(mnemonic)[1]

        # We need to coerce each opcodes operands into their
        # final `Operand` form.
        final_operands = []
        for i, operand in enumerate(operands):
            if isinstance(operand, Operand):
                # Already in Operand form.
                final_operands.append(operand)
            elif isinstance(operand, Constant):
                # Convert constants into CONSTANT_INDEX'es
                final_operands.append(Operand(
                    OperandTypes.CONSTANT_INDEX,
                    operand.index
                ))
            elif isinstance(operand, dict):
                    # lookupswitch's operand is a dict as
                # a special usability case.
                final_operands.append(operand)
            elif isinstance(operand, Label):
                final_operands.append(operand)
            else:
                # For anything else, lookup that opcode's operand
                # type from its definition.
                final_operands.append(Operand(
                    operand_fmts[i][1],
                    operand
                ))

        # Build the final, immutable `Instruction`.
        final.append(Instruction.from_mnemonic(
            mnemonic,
            operands=final_operands
        ))

    label_pcs = {}

    # The second pass, find the absolute PC for each label.
    current_pc = 0
    for ins in final:
        if isinstance(ins, Label):
            label_pcs[ins.name] = current_pc
            continue

        # size_on_disk must know the current pc because of alignment on
        # tableswitch and lookupswitch.
        current_pc += ins.size_on_disk(current_pc)

    # The third pass, now that we know where each label is we can figure
    # out the offset for each jump.
    current_pc = 0
    offset = lambda l: Operand(40, label_pcs[l.name] - current_pc)

    for ins in final:
        if isinstance(ins, Label):
            continue

        for i, operand in enumerate(ins.operands):
            if isinstance(operand, dict):
                # lookupswitch is a special case
                for k, v in operand.items():
                    if isinstance(v, Label):
                        operand[k] = offset(v)
            elif isinstance(operand, Label):
                ins.operands[i] = offset(operand)

        current_pc += ins.size_on_disk(current_pc)

        yield ins



# -*- coding: utf8 -*-
from struct import unpack, pack
from itertools import repeat
from collections import namedtuple

from jawa.attribute import Attribute, AttributeTable
from jawa.util.bytecode import (
    read_instruction,
    write_instruction
)

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO


CodeException = namedtuple('CodeException', [
    'start_pc', 'end_pc', 'handler_pc', 'catch_type'
])


class CodeAttribute(Attribute):
    """
    A `CodeAttribute` contains the executable bytecode of a single method.

    As a quick example, lets make a "HelloWorld" class with a single method
    that simply returns when it's called:

    .. code-block:: python

        from jawa import ClassFile
        from jawa.util.bytecode import Instruction

        cf = ClassFile.create('HelloWorld')

        main = cf.methods.create(
            # The name of the method
            'main',
            # The signature of the method
            '([Ljava/lang/String;)V',
            # Tell Jawa to automatically create an empty CodeAttribute for
            # us to use.
            code=True
        )
        main.code.max_locals = 1
        main.access_flags.acc_static = True
        main.code.assemble([
            Instruction.from_mnemonic('return')
        ])

        # Save it to disk so we can run it with the JVM.
        with open('HelloWorld.class', 'wb') as fout:
            cf.save(fout)

    .. note::

        Not all :class:`~jawa.methods.Method` objects will have an associated
        `CodeAttribute` - methods that are flagged as `acc_native` or
        `acc_abstract` will never have one.
    """
    ADDED_IN = '1.0.2'
    MINIMUM_CLASS_VERSION = (45, 3)

    def __init__(self, table, name_index=None):
        super(CodeAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'Code'
            ).index
        )
        self._max_stack = 0
        self._max_locals = 0
        self._ex_table = []
        self._attributes = AttributeTable(table.cf, parent=self)
        self._code = ''

    def unpack(self, info):
        """
        Read the CodeAttribute from the byte string `info`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param info: A byte string containing an unparsed CodeAttribute.
        """
        fio = StringIO(info)
        self._max_stack, self._max_locals, c_len = unpack('>HHI', fio.read(8))
        self._code = fio.read(c_len)

        # The exception table
        ex_table_len = unpack('>H', fio.read(2))[0]
        self._ex_table = []
        for _ in repeat(None, ex_table_len):
            self._ex_table.append(CodeException(
                *unpack('>HHHH', fio.read(8))
            ))
        self._attributes = AttributeTable(self._cf, parent=self)
        self._attributes.unpack(fio)
        fio.close()

    @property
    def info(self):
        """
        The `CodeAttribute` in packed byte string form.
        """
        fout = StringIO()
        fout.write(pack(
            '>HHI',
            self._max_stack,
            self._max_locals,
            len(self._code)
        ))
        fout.write(self._code)
        fout.write(pack('>H', len(self._ex_table)))
        for exception in self._ex_table:
            fout.write(pack('>HHHH', *exception))
        self._attributes.pack(fout)
        return fout.getvalue()

    @property
    def max_stack(self):
        """The maximum size of the stack."""
        return self._max_stack

    @max_stack.setter
    def max_stack(self, value):
        self._max_stack = value

    @property
    def max_locals(self):
        """The maximum number of locals."""
        return self._max_locals

    @max_locals.setter
    def max_locals(self, value):
        self._max_locals = value

    @property
    def exception_table(self):
        return self._ex_table

    @property
    def code(self):
        return self._code

    @property
    def attributes(self):
        """
        An :class:`~jawa.attribute.AttributeTable` containing all of the
        attributes associated with this `CodeAttribute`.
        """
        return self._attributes

    def assemble(self, code):
        """
        Assembles an iterable of :class:`~jawa.util.bytecode.Instruction`
        objects into a method's code body.
        """
        fout = StringIO()
        for ins in code:
            write_instruction(fout, fout.tell(), ins)
        self._code = fout.getvalue()
        fout.close()

    def disassemble(self):
        """
        Disassembles this method, yielding an iterable of
        :class:`~jawa.util.bytecode.Instruction` objects.
        """
        fio = StringIO(self._code)

        for ins in iter(lambda: read_instruction(fio, fio.tell()), None):
            yield ins



# -*- coding: utf8 -*-
__all__ = (
    'ConstantPool',
    'Constant',
    'ConstantUTF8',
    'ConstantClass',
    'ConstantString',
    'ConstantRef',
    'ConstantFieldRef',
    'ConstantMethodRef',
    'ConstantInterfaceMethodRef',
    'ConstantInteger',
    'ConstantFloat',
    'ConstantLong',
    'ConstantDouble',
    'ConstantNameAndType',
    'ConstantMethodHandle',
    'ConstantMethodType',
    'ConstantInvokeDynamic',
)

from struct import unpack, pack

from jawa.util.utf import decode_modified_utf8, encode_modified_utf8


class Constant(object):
    """
    The base class for all ``Constant*`` types.
    """
    def __init__(self, pool, index):
        self._pool = pool
        self._index = index

    @property
    def pool(self):
        """
        The constant pool this constant belongs to.
        """
        return self._pool

    @property
    def index(self):
        """
        The index of this constant in the constant pool.
        """
        return self._index


class ConstantNumber(Constant):
    def __init__(self, pool, index, value):
        super(ConstantNumber, self).__init__(pool, index)
        self.value = value

    def __repr__(self):
        return '{0}(value={1!r})'.format(self.__class__.__name__, self.value)


class ConstantUTF8(Constant):
    TAG = 1

    def __init__(self, pool, index, value):
        super(ConstantUTF8, self).__init__(pool, index)
        self.value = value

    def __repr__(self):
        return 'ConstantUTF8(value={0!r})'.format(self.value)


class ConstantInteger(ConstantNumber):
    TAG = 3


class ConstantFloat(ConstantNumber):
    TAG = 4


class ConstantLong(ConstantNumber):
    TAG = 5


class ConstantDouble(ConstantNumber):
    TAG = 6


class ConstantClass(Constant):
    TAG = 7

    def __init__(self, pool, index, name_index):
        super(ConstantClass, self).__init__(pool, index)
        self._name_index = name_index

    @property
    def name(self):
        return self.pool.get(self._name_index)

    def __repr__(self):
        return 'ConstantClass(name={0!r})'.format(self.name)


class ConstantString(Constant):
    TAG = 8

    def __init__(self, pool, index, string_index):
        super(ConstantString, self).__init__(pool, index)
        self._string_index = string_index

    @property
    def string(self):
        return self.pool.get(self._string_index)

    def __repr__(self):
        return 'ConstantString(string={0!r})'.format(self.string)


class ConstantRef(Constant):
    def __init__(self, pool, index, class_index, name_and_type_index):
        super(ConstantRef, self).__init__(pool, index)
        self._class_index = class_index
        self._name_and_type_index = name_and_type_index

    @property
    def class_(self):
        return self.pool.get(self._class_index)

    @property
    def name_and_type(self):
        return self.pool.get(self._name_and_type_index)

    def __repr__(self):
        return '{0}(class_={1!r}, name_and_type={2!r})'.format(
            self.__class__.__name__, self.class_, self.name_and_type)


class ConstantFieldRef(ConstantRef):
    TAG = 9


class ConstantMethodRef(ConstantRef):
    TAG = 10


class ConstantInterfaceMethodRef(ConstantRef):
    TAG = 11


class ConstantNameAndType(Constant):
    TAG = 12

    def __init__(self, pool, index, name_index, descriptor_index):
        super(ConstantNameAndType, self).__init__(pool, index)
        self._name_index = name_index
        self._descriptor_index = descriptor_index

    @property
    def name(self):
        return self.pool.get(self._name_index)

    @property
    def descriptor(self):
        return self.pool.get(self._descriptor_index)

    def __repr__(self):
        return 'ConstantNameAndType(name={0!r}, descriptor={1!r})'.format(
            self.name, self.descriptor)

class ConstantMethodHandle(Constant):
    TAG = 15

    def __init__(self, pool, index, reference_kind, reference_index):
        super(ConstantMethodHandle, self).__init__(pool, index)
        self._reference_kind = reference_kind
        self._reference_index = reference_index

    @property
    def kind(self):
        return self.pool.get(self._reference_kind)

    @property
    def reference(self):
        return self.pool.get(self._reference_index)

    def __repr__(self):
        return 'ConstantMethodHandle(kind={0!r}, index={1!r})'.format(
            self.kind, self.reference)

class ConstantMethodType(Constant):
    TAG = 16

    def __init__(self, pool, index, descriptor_index):
        super(ConstantMethodType, self).__init__(pool, index)
        self._descriptor_index = descriptor_index

    @property
    def descriptor(self):
        return self.pool.get(self._descriptor_index)

    def __repr__(self):
        return 'ConstantMethodType(descriptor={0!r})'.format(
            self.descriptor)

class ConstantInvokeDynamic(Constant):
    TAG = 18

    def __init__(self, pool, index, bootstrap_method_attr_index, name_and_type_index):
        super(ConstantInvokeDynamic, self).__init__(pool, index)
        self._bootstrap_method_attr_index = bootstrap_method_attr_index
        self._name_and_type_index = name_and_type_index

    @property
    def methodAttr(self):
        return self.pool.get(self._bootstrap_method_attr_index)

    @property
    def nameAndType(self):
        return self.pool.get(self._name_and_type_index)

    def __repr__(self):
        return 'ConstantInvokeDynamic(methodAttr={0!r}, nameAndType={1!r})'.format(
            self.methodAttr, self.nameAndType)

_constant_types = (
    None,
    ConstantUTF8,
    None,
    ConstantInteger,
    ConstantFloat,
    ConstantLong,
    ConstantDouble,
    ConstantClass,
    ConstantString,
    ConstantFieldRef,
    ConstantMethodRef,
    ConstantInterfaceMethodRef,
    ConstantNameAndType,
    None,
    None,
    ConstantMethodHandle,
    ConstantMethodType,
    None,
    ConstantInvokeDynamic
)


# The format and size-on-disk of each type of constant
# in the constant pool.
_constant_fmts = (
    None, None, None,
    ('>i', 4),
    ('>f', 4),
    ('>q', 8),
    ('>d', 8),
    ('>H', 2),
    ('>H', 2),
    ('>HH', 4),
    ('>HH', 4),
    ('>HH', 4),
    ('>HH', 4),
    None,
    None,
    ('>BH', 3),
    ('>H', 2),
    None,
    ('>HH', 4)
)


class ConstantPool(object):
    def __init__(self):
        self._pool = [None]

    def append(self, constant):
        """
        Appends a new constant to the end of the pool.
        """
        self._pool.append(constant)

    def __iter__(self):
        for index, constant in enumerate(self._pool):
            if constant is not None:
                yield self.get(index)

    def get(self, index):
        """
        Returns the `Constant` at `index`, raising a KeyError if it
        does not exist.
        """
        constant = self._pool[index]
        if not isinstance(constant, Constant):
            constant = _constant_types[constant[0]](self, index, *constant[1:])
            self._pool[index] = constant
        return constant

    def __getitem__(self, idx):
        return self.get(idx)

    def __setitem__(self, idx, value):
        self._pool[idx] = value

    def find(self, type_=None, f=None):
        """
        Iterates over the pool, yielding each matching ``Constant``. Calling
        without any arguments is equivelent to iterating over the pool.

        :param type_: Any subclass of :class:`Constant` or ``None``.
        :param f: Any callable which takes one argument (the constant).
        """
        for constant in self:
            if type_ is not None and not isinstance(constant, type_):
                continue

            if f is not None and not f(constant):
                continue

            yield constant

    def find_one(self, *args, **kwargs):
        """
        Same as ``find()`` but returns only the first result, or `None` if
        nothing was found.
        """
        try:
            return next(self.find(*args, **kwargs))
        except StopIteration:
            return None

    def create_utf8(self, value):
        """
        Creates a new :class:`ConstantUTF8`, adding it to the pool and
        returning it.

        :param value: The value of the new UTF8 string.
        """
        self.append((1, unicode(value)))
        return self.get(self.raw_count - 1)

    def create_integer(self, value):
        """
        Creates a new :class:`ConstantInteger`, adding it to the pool and
        returning it.

        :param value: The value of the new integer.
        """
        self.append((3, value))
        return self.get(self.raw_count - 1)

    def create_float(self, value):
        """
        Creates a new :class:`ConstantFloat`, adding it to the pool and
        returning it.

        :param value: The value of the new float.
        """
        self.append((4, value))
        return self.get(self.raw_count - 1)

    def create_long(self, value):
        """
        Creates a new :class:`ConstantLong`, adding it to the pool and
        returning it.

        :param value: The value of the new long.
        """
        self.append((5, value))
        self.append(None)
        return self.get(self.raw_count - 2)

    def create_double(self, value):
        """
        Creates a new :class:`ConstantDouble`, adding it to the pool and
        returning it.

        :param value: The value of the new Double.
        """
        self.append((6, value))
        self.append(None)
        return self.get(self.raw_count - 2)

    def create_class(self, name):
        """
        Creates a new :class:`ConstantClass`, adding it to the pool and
        returning it.

        :param name: The name of the new class.
        """
        self.append((
            7,
            self.create_utf8(name).index
        ))
        return self.get(self.raw_count - 1)

    def create_string(self, value):
        """
        Creates a new :class:`ConstantString`, adding it to the pool and
        returning it.

        :param value: The value of the new string as a UTF8 string.
        """
        self.append((
            8,
            self.create_utf8(value).index
        ))
        return self.get(self.raw_count - 1)

    def create_name_and_type(self, name, descriptor):
        """
        Creates a new :class:`ConstantNameAndType`, adding it to the pool and
        returning it.

        :param name: The name of the class.
        :param descriptor: The descriptor for `name`.
        """
        self.append((
            12,
            self.create_utf8(name).index,
            self.create_utf8(descriptor).index
        ))
        return self.get(self.raw_count - 1)

    def create_field_ref(self, class_, field, descriptor):
        """
        Creates a new :class:`ConstantFieldRef`, adding it to the pool and
        returning it.

        :param class_: The name of the class to which `field` belongs.
        :param field: The name of the field.
        :param descriptor: The descriptor for `field`.
        """
        self.append((
            9,
            self.create_class(class_).index,
            self.create_name_and_type(field, descriptor).index
        ))
        return self.get(self.raw_count - 1)

    def create_method_ref(self, class_, method, descriptor):
        """
        Creates a new :class:`ConstantMethodRef`, adding it to the pool and
        returning it.

        :param class_: The name of the class to which `method` belongs.
        :param method: The name of the method.
        :param descriptor: The descriptor for `method`.
        """
        self.append((
            10,
            self.create_class(class_).index,
            self.create_name_and_type(method, descriptor).index
        ))
        return self.get(self.raw_count - 1)

    def create_interface_method_ref(self, class_, if_method, descriptor):
        """
        Creates a new :class:`ConstantInterfaceMethodRef`, adding it to the
        pool and returning it.

        :param class_: The name of the class to which `if_method` belongs.
        :param if_method: The name of the interface method.
        :param descriptor: The descriptor for `if_method`.
        """
        self.append((
            11,
            self.create_class(class_).index,
            self.create_name_and_type(if_method, descriptor).index
        ))
        return self.get(self.raw_count - 1)

    def unpack(self, fio):
        """
        Read the ConstantPool from the file-like object `fio`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param fio: Any file-like object providing `read()`
        """
        # Reads in the ConstantPool (constant_pool in the JVM Spec)
        constant_pool_count = unpack('>H', fio.read(2))[0]

        # Pull this locally so CPython doesn't do a lookup each time.
        read = fio.read

        while constant_pool_count > 1:
            constant_pool_count -= 1
            # The 1-byte prefix identifies the type of constant.
            tag = unpack('>B', read(1))[0]

            if tag == 1:
                # CONSTANT_Utf8_info, a length prefixed UTF-8-ish string.
                length = unpack('>H', read(2))[0]
                self.append((tag, decode_modified_utf8(read(length))))
            else:
                # Every other constant type is trivial.
                fmt, size = _constant_fmts[tag]
                self.append((tag,) + unpack(fmt, read(size)))
                if tag in (5, 6):
                    # LONG (5) and DOUBLE (6) count as two entries in the
                    # pool.
                    self.append(None)
                    constant_pool_count -= 1

    def pack(self, fout):
        """
        Write the ConstantPool to the file-like object `fout`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be calle=d for you when saving a ClassFile.

        :param fout: Any file-like object providing `write()`
        """
        write = fout.write
        write(pack('>H', self.raw_count))

        for constant in self:
            if isinstance(constant, ConstantUTF8):
                encoded_value = encode_modified_utf8(constant.value)
                length = len(encoded_value)
                write(pack(
                    '>BH',
                    constant.TAG,
                    length
                ))
                write(encoded_value)
            elif isinstance(constant, ConstantInteger):
                write(pack(
                    '>Bi',
                    constant.TAG,
                    constant.value
                ))
            elif isinstance(constant, ConstantFloat):
                write(pack(
                    '>Bf',
                    constant.TAG,
                    constant.value
                ))
            elif isinstance(constant, ConstantLong):
                write(pack(
                    '>Bq',
                    constant.TAG,
                    constant.value
                ))
            elif isinstance(constant, ConstantDouble):
                write(pack(
                    '>Bd',
                    constant.TAG,
                    constant.value
                ))
            elif isinstance(constant, ConstantClass):
                write(pack(
                    '>BH',
                    constant.TAG,
                    constant._name_index
                ))
            elif isinstance(constant, ConstantString):
                write(pack(
                    '>BH',
                    constant.TAG,
                    constant._string_index
                ))
            elif isinstance(constant, ConstantRef):
                write(pack(
                    '>BHH',
                    constant.TAG,
                    constant._class_index,
                    constant._name_and_type_index
                ))
            elif isinstance(constant, ConstantNameAndType):
                write(pack(
                    '>BHH',
                    constant.TAG,
                    constant._name_index,
                    constant._descriptor_index
                ))
            elif isinstance(constant, ConstantMethodHandle):
                write(pack(
                    '>BBH',
                    constant.TAG,
                    constant._reference_kind,
                    constant._reference_index
                ))
            elif isinstance(constant, ConstantMethodType):
                write(pack(
                    '>BH',
                    constant.TAG,
                    constant._descriptor_index
                ))
            elif isinstance(constant, ConstantInvokeDynamic):
                write(pack(
                    '>BHH',
                    constant.TAG,
                    constant._bootstrap_method_attr_index,
                    constant._name_and_type_index
                ))

    def __len__(self):
        """
        The number of `Constants` in the `ConstantPool`, excluding padding.
        """
        count = 0
        for constant in self._pool:
            if constant is not None:
                count += 1
        return count

    @property
    def raw_count(self):
        """
        The number of `Constants` in the `ConstantPool`, including padding.
        """
        return len(self._pool)



# -*- coding: utf8 -*-
from struct import unpack, pack
from itertools import repeat


class Attribute(object):
    def __init__(self, table, name_index=None):
        self._table = table
        self._cf = table.cf
        self._name_index = name_index

    @property
    def info(self):
        """
        This attribute packed into its on-disk representation.
        """
        raise NotImplementedError()

    @property
    def name(self):
        """
        The :class:`~jawa.constants.ConstantUtf8` with the name of this
        attribute.
        """
        return self._cf.constants[self._name_index]

    @property
    def table(self):
        """
        The AttributeTable that owns this attribute, if any.
        """
        return self._table

    @property
    def cf(self):
        """
        The ClassFile that owns this attribute, if any.
        """
        return self._cf

    def unpack(self, info):
        """
        Parses an instance of this attribute from the blob `info`.
        """
        raise NotImplementedError()

    def pack(self):
        """
        This attribute packed into its on-disk representation.
        """
        return self.info


class UnknownAttribute(Attribute):
    def unpack(self, info):
        self._info = info

    @property
    def info(self):
        return self._info


class AttributeTable(object):
    def __init__(self, cf, parent=None):
        self._cf = cf
        self._table = []
        self._parent = parent

    def _get_type(self, name):
        """
        Returns a `Attribute` subclass that can handle the attribute
        of type `name`, or :class:~jawa.attribute.UnknownAttribute if
        none is found.
        """
        return default_parsers.get(name, UnknownAttribute)

    def unpack(self, fio):
        """
        Read the ConstantPool from the file-like object `fio`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param fio: Any file-like object providing `read()`
        """
        count = unpack('>H', fio.read(2))[0]
        for _ in repeat(None, count):
            name_index, length = unpack('>HI', fio.read(6))
            name = self._cf.constants[name_index].value

            type_ = self._get_type(name)
            attribute = type_(self, name_index=name_index)

            attribute.unpack(fio.read(length))
            self._table.append(attribute)

    def pack(self, fout):
        """
        Write the AttributeTable to the file-like object `fout`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when saving a ClassFile.

        :param fout: Any file-like object providing `write()`
        """
        fout.write(pack('>H', len(self)))
        for attribute in self._table:
            info = attribute.info
            fout.write(pack(
                '>HI',
                attribute.name.index,
                len(info)
            ))
            fout.write(info)

    def create(self, type_, *args, **kwargs):
        """
        Creates a new attribute of `type_`, appending it to the attribute
        table and returning it.
        """
        attribute = type_(self, *args, **kwargs)
        self.append(attribute)
        return attribute

    def __len__(self):
        return len(self._table)

    def append(self, attribute):
        self._table.append(attribute)

    def find(self, name=None, f=None):
        for attribute in self._table:
            if name is not None and not attribute.name.value == name:
                continue

            if f is not None and not f(attribute):
                continue

            yield attribute

    def find_one(self, *args, **kwargs):
        """
        Same as ``find()`` but returns only the first result, or `None` if
        nothing was found.
        """
        try:
            return next(self.find(*args, **kwargs))
        except StopIteration:
            return None

    @property
    def cf(self):
        """
        The ClassFile that owns this Attribute, if any.
        """
        return self._cf

    @property
    def parent(self):
        """
        The parent attribute, if any.

        If this AttributeTable belongs to another Attribute, this will
        reference that attribute. For example, when parsing a StackMapTable
        attribute, this would point to the owning Code attribute.
        """
        return self._parent


# Attributes can contain other attributes and AttributeTable's,
# thus we have to do our import here.
# pylint: disable=cyclic-import
from jawa.attributes.code import CodeAttribute
from jawa.attributes.source_file import SourceFileAttribute
from jawa.attributes.constant_value import ConstantValueAttribute

default_parsers = {
    'Code': CodeAttribute,
    'SourceFile': SourceFileAttribute,
    'ConstantValue': ConstantValueAttribute
}



# -*- coding: utf8 -*-
from jawa.cf import *
from jawa.fields import *
from jawa.methods import *
from jawa.constants import *
from jawa.attribute import *
from jawa.attributes.code import *
from jawa.attributes.source_file import *
from jawa.attributes.constant_value import *
